@@ -40,7 +40,7 @@
 
 import numpy as np
 import scipy.linalg
-from scipy.special import gammaln
+from scipy.special import gamma
 
 
 def normalize_row_vectors(x):
@@ -104,11 +104,11 @@ def generate_sources(rho, N, T, K):
     return S_matrix
 
 
-def MGGD_generation(N, p=None, correlation_structure=None, rho=None, beta=1, Sigma=None):
+def MGGD_generation(N, dim=None, correlation_structure=None, rho=None, beta=1, cov=None):
     """
     Contains six functions that can be used to generate multivariate generalized Gaussian
-    distributed (MGGD) sources. If Sigma is given, none of the functions will be used but data will
-    be generated using Sigma.
+    distributed (MGGD) sources. If cov is given, none of the functions will be used but data will
+    be generated using cov.
     
     
     Parameters
@@ -116,11 +116,11 @@ def MGGD_generation(N, p=None, correlation_structure=None, rho=None, beta=1, Sig
     N : int
         Number of realizations / samples
         
-    p : int, optional
-        Dimension / number of variables. Must be at least 2. Must be given if Sigma is None.
+    dim : int, optional
+        Dimension / number of variables. Must be given if cov is None.
 
     correlation_structure : str, optional
-        Which structure of correlation matrix to use. Must be given if Sigma is None.
+        Which structure of correlation matrix to use. Must be given if cov is None.
             - 'uniform': elements are equal to rho
             - 'ar' (auto regressive): cov[i, j] = rho ** (np.abs(i - j))
             - 'q_qt': normalized random samples ~ U[-0.2, 0.8) multiplied by themselves
@@ -130,12 +130,12 @@ def MGGD_generation(N, p=None, correlation_structure=None, rho=None, beta=1, Sig
                        block-elements have same value, diagonal elements are 1
         
     rho : float or dict or np.ndarray or list, optional
-        Correlation value. Must be given if Sigma is None and correlation_structure is not 'q_qt'.
+        Correlation value. Must be given if cov is None and correlation_structure is not 'q_qt'.
         If correlation_structure=='two_rho', rho must be a dictionary with keys 'val' and 'idx'.
             'val' is a np.ndarray or list with two values of correlation: 'val[0]' provides the
             basic correlation.
             'idx' is a np.ndarray or list that gives the specific locations where 'val[1]' is set.
-        If correlation_structure=='rho_list', rho must be a np.ndarray or list with length p-1.
+        If correlation_structure=='rho_list', rho must be a np.ndarray or list with length dim-1.
         If correlation_structure=='block', rho is a dict with keys 'blocks' and 'val'.
             - rho['blocks'] must be tuple or list of tuples, where each tuple contains
               (correlation value of the block, start idx of the block, length of the blocks).
@@ -148,16 +148,16 @@ def MGGD_generation(N, p=None, correlation_structure=None, rho=None, beta=1, Sig
               tails.
             - If beta is greater than 1, it is less peaky with lighter tails.
 
-    Sigma : np.ndarray, optional
-        User-defined covariance matrix of dimensions p x p, where p is the number of sources
+    cov : np.ndarray, optional
+        User-defined covariance matrix of dimensions dim x dim, where dim is the number of sources
 
     Returns
     -------
     X : np.ndarray
-        matrix of dimensions p x N, which contains N samples of each of the p MGGD sources
+        matrix of dimensions dim x N, which contains N samples of each of the dim MGGD sources
 
-    COV: np.ndarray
-        covariance matrix
+    cov: np.ndarray
+        covariance matrix of X (either generated by this function, or same as input)
         
     
     Notes
@@ -179,17 +179,17 @@ def MGGD_generation(N, p=None, correlation_structure=None, rho=None, beta=1, Sig
     if type(N) is not int or N < 2:
         raise AssertionError("'N' must be an integer and must be at least 2.")
 
-    if Sigma is None:
+    if cov is None:
 
-        if p is None:
-            raise AssertionError("Either 'p' or 'Sigma' must be given.")
+        if dim is None:
+            raise AssertionError("Either 'dim' or 'cov' must be given.")
         elif correlation_structure is None:
-            raise AssertionError("Either 'correlation_structure' or 'Sigma' must be given.")
+            raise AssertionError("Either 'correlation_structure' or 'cov' must be given.")
         elif rho is None and correlation_structure != 'q_qt':
-            raise AssertionError("Either 'rho' or 'Sigma' must be given.")
+            raise AssertionError("Either 'rho' or 'cov' must be given.")
 
-        if type(p) is not int or p < 2:
-            raise AssertionError("'p' must be an integer and must be at least 2.")
+        if type(dim) is not int:
+            raise AssertionError("'dim' must be an integer.")
 
         if correlation_structure == 'two_rho':
             if type(rho) is not dict:
@@ -223,81 +223,80 @@ def MGGD_generation(N, p=None, correlation_structure=None, rho=None, beta=1, Sig
             if type(rho) is not float and type(rho) is not int:
                 raise AssertionError("'rho' must be a float or int value.")
 
-        if correlation_structure not in ['uniform', 'ar', 'q_qt', 'two_rho', 'rho_list', 'block']:
-            raise AssertionError("'correlation_structure' must be 'uniform', 'ar', 'q_qt', "
-                                 "'two_rho', 'rho_list', or 'block'.")
-
     else:
-        p = Sigma.shape[0]
+        dim = cov.shape[0]
 
-        if Sigma.ndim != 2:
-            raise AssertionError("'Sigma' must be a 2D matrix.")
-        elif Sigma.shape[0] != Sigma.shape[1]:
-            raise AssertionError("Sigma' must be a square matrix.")
+        if cov.ndim != 2:
+            raise AssertionError("'cov' must be a 2D matrix.")
+        elif cov.shape[0] != cov.shape[1]:
+            raise AssertionError("cov' must be a square matrix.")
+        try:
+            np.linalg.cholesky(cov)
+        except:
+            raise AssertionError("Covariance matrix must be positive semidefinite.")
 
-    # ---------Q. Long 02/2018--------########
-    betah = 2 * beta
-    etanln = 1 / beta * np.log(2) + gammaln((p + 2) / betah) - gammaln(p / betah) - np.log(p)
-    etan = np.exp(etanln)
-    # ---------Q. Long 02/2018--------########
+    if cov is None:
+        # This is the covariance matrix of X
+        cov = np.abs(_create_covariance_matrix(dim, correlation_structure, rho))
 
-    X = _random_sphere(N, p).T
+        # make sure randomly generated cov is positive semidefinite. If not, regenerate
+        flag = 0
+        while flag == 0:
+            try:
+                np.linalg.cholesky(cov)
+                flag = 1
+            except:
+                cov = np.abs(_create_covariance_matrix(dim, correlation_structure, rho))
 
-    if Sigma is None:
-        # This is the covariance matrix we want
-        COV1 = np.abs(_create_covariance_matrix(p, correlation_structure, rho))
-    else:
-        COV1 = np.copy(Sigma)
+    u = _random_sphere(N, dim).T  # u^(n) in (3)
 
-    COV = COV1 / etan  # Use this as the Sigma in Gomez's paper
+    # X is parametrized by scaled version of cov, see proposition 3.2 (iii)
+    scaling = 2 ** (1 / beta) * gamma((dim + 2) / (2 * beta)) / (dim * gamma(dim / (2 * beta)))
+    Sigma = cov / scaling
 
-    # check if covariance matrix is positive semidefintive
-    try:
-        # Use Cholesky decomposition instead of the square root of Sigma
-        X = np.linalg.cholesky(COV) @ X
-    except:
-        raise AssertionError("Covariance matrix is not positive semidefinite.")
+    # chol(Sigma) = A^T in (3)
+    u_A = np.linalg.cholesky(Sigma) @ u
 
-    tau = np.random.gamma(p / (2 * beta), scale=2, size=N) ** (1 / (2 * beta))
+    R = np.random.gamma(dim / (2 * beta), scale=2, size=N) ** (1 / (2 * beta))  # R in (3) with density shown in (4)
 
-    X = tau * X
+    X = R * u_A
 
-    return X, COV
+    return X, cov
 
 
-def _create_covariance_matrix(p, correlation_structure, rho):
+def _create_covariance_matrix(dim, correlation_structure, rho):
     if correlation_structure == 'uniform':
-        Sigma = np.full((p, p), rho)
+        Sigma = np.full((dim, dim), rho)
         np.fill_diagonal(Sigma, 1)
 
     elif correlation_structure == 'ar':
-        Sigma = np.eye(p)
-        for i in range(p):
-            for j in range(p):
+        Sigma = np.eye(dim)
+        for i in range(dim):
+            for j in range(dim):
                 Sigma[i, j] = rho ** (np.abs(i - j))
 
     elif correlation_structure == 'q_qt':
-        R = normalize_row_vectors(np.random.rand(p, p) - 0.2)
+        R = normalize_row_vectors(np.random.rand(dim, dim) - 0.2)
         Sigma = R @ R.T
 
     elif correlation_structure == 'two_rho':
-        Sigma = np.full((p, p), rho['val'][0])
+        Sigma = np.full((dim, dim), rho['val'][0])
         for i in rho['idx']:
             for j in rho['idx']:
                 Sigma[i, j] = rho['val'][1]
         np.fill_diagonal(Sigma, 1)
 
     elif correlation_structure == 'rho_list':
-        Sigma = np.eye(p)
+        Sigma = np.eye(dim)
         n = len(rho)
-        if np.mod(p - 1, n):
-            raise AssertionError("'rho' must contain 'p'-1 elements")
-        M = (p - 1) // n
+        if np.mod(dim - 1, n):
+            raise AssertionError("'rho' must contain 'dim'-1 elements")
+        M = (dim - 1) // n
         for j in range(n):
             Sigma[0, M * j + 1:M * (j + 1) + 1] = rho[j] * np.ones(M)
             Sigma[M * j + 1:M * (j + 1) + 1, 0] = rho[j] * np.ones(M)
-        for i in range(1, p):
-            for j in range(i + 1, p):
+        for i in range(1, dim):
+            for j in range(i + 1, dim):
                 Sigma[i, j] = Sigma[i - 1, j]
                 Sigma[j, i] = Sigma[i, j]
 
@@ -307,11 +306,15 @@ def _create_covariance_matrix(p, correlation_structure, rho):
         else:
             csl = rho['blocks']
 
-        Sigma = np.full((p, p), rho['val'])
+        Sigma = np.full((dim, dim), rho['val'])
         for (correlation, start, length) in csl:
             Sigma[start:start + length, start:start + length] = correlation
         np.fill_diagonal(Sigma, 1)
 
+    else:
+        raise AssertionError("'correlation_structure' must be 'uniform', 'ar', 'q_qt', "
+                             "'two_rho', 'rho_list', or 'block'.")
+
     return Sigma
 
 
